package dev.cache.infinispan;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.concurrent.Future;


import org.infinispan.Cache;
import org.infinispan.configuration.cache.CacheMode;
import org.infinispan.configuration.cache.Configuration;
import org.infinispan.configuration.cache.ConfigurationBuilder;
import org.infinispan.configuration.global.GlobalConfigurationBuilder;
import org.infinispan.distexec.DefaultExecutorService;
import org.infinispan.distexec.DistributedCallable;
import org.infinispan.distexec.DistributedExecutorService;
import org.infinispan.eviction.EvictionStrategy;
import org.infinispan.manager.DefaultCacheManager;
import org.infinispan.notifications.Listener;
import org.infinispan.notifications.cachelistener.annotation.CacheEntriesEvicted;
import org.infinispan.notifications.cachelistener.annotation.CacheEntryCreated;
import org.infinispan.notifications.cachelistener.annotation.CacheEntryModified;
import org.infinispan.notifications.cachelistener.annotation.CacheEntryRemoved;
import org.infinispan.notifications.cachelistener.event.CacheEntriesEvictedEvent;
import org.infinispan.notifications.cachelistener.event.CacheEntryCreatedEvent;
import org.infinispan.notifications.cachelistener.event.CacheEntryModifiedEvent;
import org.infinispan.notifications.cachelistener.event.CacheEntryRemovedEvent;
import org.infinispan.remoting.transport.Address;



/**
 * 
 * @author dudalov
 *
 * java -cp ".;build/classes;lib/*"  [options]    dev.cache.infinispan.SimpleInteractiveTest
 * where options:
 *     --listener   if you want to run with the listener that print events to the console
 *     
 *  Change classpath accordingly to point to:
 *    build/classes   .... location of class-file for this class
 *    lib ................ folder with libraries from infinispan-5.1.2.FINAL
 */
public class SimpleInteractiveTest  {

	private static String CLASSNAME = SimpleInteractiveTest.class.getCanonicalName();
	private DefaultCacheManager _manager;
	private static boolean useListener = false;

	
	/**
	 * List of supported commands
	 */
	private enum COMMAND {
		ADD("Add a Resource to the CURRENT cache"), 
		GET("Get list of the Resources with the specified name"), 
		LIST("List all Resources"), 
		TRIGGER("Imitate changes of the Resource with a name"),
		EXECUTE("Execute predefined Callable everywhere in the cluster"),
		QUIT("Quit this test"), 
		HELP("Print this help");
		
		private final String description;
		COMMAND(String description) {
			this.description = description;
		}
		public String getDescription() { return description; } 
	}
	
	private static void log(String s) {
		System.out.println(s);
	}

	private static void debug(String method, String msg) {
		System.out.format("[%s] %s : %s : %s\n", CLASSNAME, new Date(), method, msg);
	}

	
	public SimpleInteractiveTest() {
		_manager = new DefaultCacheManager(
	            GlobalConfigurationBuilder.defaultClusteredBuilder()
	            .globalJmxStatistics().enable()
                .transport().addProperty("configurationFile", "jgroups-udp.xml")
                .build(),
                new ConfigurationBuilder()
                	.clustering().cacheMode(CacheMode.REPL_SYNC)
                	.build()
				);		
	}


	protected void terminate() throws Exception {
		_manager.stop();
	}

	private Cache<String, String> getCache(String cacheName, int capacity) {
		EvictionStrategy evictionStrategy = EvictionStrategy.LRU;

		Configuration config = new ConfigurationBuilder().eviction()
				.maxEntries(capacity).strategy(evictionStrategy).expiration()
				.wakeUpInterval(5000L).maxIdle(120000L)
				.clustering().cacheMode(CacheMode.REPL_SYNC)
				.build();

		_manager.defineConfiguration(cacheName, config);
		Cache<String, String> infinispanCache = _manager.getCache(cacheName);
		return infinispanCache;
	}
	
	private List<String> getAllResources(Cache<String, String> cache) {
		List<String> resources = new ArrayList<String>();
		Set<java.util.Map.Entry<String, String>> entries = cache.entrySet();
		if ( null != entries ) {
			for ( java.util.Map.Entry<String, String> entry : entries) {
				String strValue = entry.getValue();
				if ( null != strValue ) {
					resources.add(strValue);
				}
			}
		}
		return resources;
	}
	
	private String getPrompt(String cacheName) {
		if ( null == cacheName ) {
			return "> ";
		}
		String prompt = String.format("%s> ", cacheName);
		return prompt;
	}
	
	public void start() throws Exception {
		
		String cacheName = "FooLookupCache";
		int cacheSize = 64;
		
		log( String.format("%s Now lets' create cache \"%s\"", new Date(), cacheName));
		Cache<String, String> cache = getCache(cacheName, cacheSize); 
		
	
		// add a listenter
		EntryChangeListener listener = new EntryChangeListener();
		if ( useListener ) {
			cache.addListener(listener);
		}
		 
        IOUtils.dumpWelcomeMessage();
        
        String rsName = null;
        String resource = null;
        
		List<String> resources = null;
        
        COMMAND command = null;
        while (true)  {
        	String[] cmdline = null;
        	try {
        		String prompt = getPrompt(cacheName);
        		String input = IOUtils.readLine(prompt).toUpperCase().trim();
        		if ( input.length() <= 1 ) {
        			continue;
        		}
        		cmdline = input.split("\\s", 2);
        		
                command = COMMAND.valueOf( cmdline[0] );
        	}  catch (Exception e) {
        		//e.printStackTrace();
        		log("Unsupported command. Use one of: " + Arrays.asList(COMMAND.values()));
        		continue;
        	}
             
            switch (command) {
            case ADD:
            	if ( cmdline.length > 1) {
                	rsName = cmdline[1];
            	}  else {
                	rsName = IOUtils.readLine("Enter resource name: ");
            	}
            	cache.put(rsName, rsName);
            	break;
            case GET:
            	if ( cmdline.length > 1) {
                	rsName = cmdline[1];
            	}  else {
                	rsName = IOUtils.readLine("Enter resource name: ");
            	}
            	resource = cache.get(rsName);
            	if ( null == resource ) {
            		log(String.format("There are resources by the name \"%s\"", rsName));
            	}  else {
            		log(String.format("  %s", resource));
            	}
            	break;
            case TRIGGER:
            	// imitates Resource changes
            	if ( cmdline.length > 1) {
                	rsName = cmdline[1];
            	}  else {
                	rsName = IOUtils.readLine("Enter resource name to update: ");
            	}
            	resource = cache.get(rsName);
            	if ( null != resource ) {
        			log(String.format("  updating %s ...", resource));
        			cache.remove(resource);
            		// and add them back
        			cache.put(resource, resource);
            	}
            	break;
            case EXECUTE:
            	try {
            		resources = getAllResources(cache);
            		Set<String> keys = new HashSet<String>();
            		for (String next : resources ) {
        				keys.add( next );
        				//break;  // one key only
            		}
                	DistributedExecutorService executor = new DefaultExecutorService(cache);
                	Object[] keyArray = keys.toArray();
                	log(" Submitting tasks for: " + Arrays.toString(keyArray));
                	List<Future<String>> tasks = executor.submitEverywhere(new TestTask(), keyArray);
                	// wait right here for the results
                	for (Future<String> task : tasks ) {
                		try {
                    		String result = task.get();
                    		log("\n***  Result: " + result);
                    	}  catch (Exception e) {
                    		log("\n***  Failed to get result. Exception: " + e);
                    		e.printStackTrace();
                    	}
                	}
            	}  catch (Exception e) {
            		log("Failed to execute task. Exception: " + e);
            		e.printStackTrace();
            	}
            	
            	break;
            case LIST:
            	// ResourceBase.toString returns getDisplayName()
            	resources = getAllResources(cache);
            	log(String.format("Found %s resources:", resources.size()));
            	for (String next : resources ) {
            		log(String.format("  %s", next));
            	}
            	break;
            case QUIT:
            	_manager.stop();
            	System.exit(0);
                return;
            case HELP:
            	log("Supported commands: " + Arrays.asList(COMMAND.values()));
            	for (COMMAND next : COMMAND.values()) {
            		log(String.format("  %10s ..... %s", next, next.getDescription()));
            	}
            	break;
            }        
        }
	}

	/**
	 * 
	 * Inner classes
	 *
	 */
	private static class TestTask implements DistributedCallable<String, String, String>, Serializable {

		private static final long serialVersionUID = 3752902313615389539L;
		
		private Cache<String, String> cache;
		private Set<String> inputKeys;
		
		@Override
		public String call() throws Exception {
			final String SEPARATOR = ", ";
			
			Address localAddress = null;
			try {
				localAddress = this.cache.getCacheManager().getAddress();
			}  catch (Exception e) {
				// not critical for now
				log("Problems to get local Address: " + e);
			}
			
			StringBuilder sb = new StringBuilder("[TestTask (on node=").append(localAddress).append(")] for ");
			if ( null == inputKeys || inputKeys.isEmpty() ) {
				sb.append("no keys");
			}  else {
				sb.append(inputKeys.size()).append(" keys for Resources:\n");
				String separator = "";
				for (String key : inputKeys ) {
					sb.append(separator).append(key);
					separator = SEPARATOR;
				}
			}
			String result = sb.toString();
			log("" + new Date() + "  Executed the task. Result: " + result);
			return result;
		}

		@Override
		public void setEnvironment(Cache<String, String> cache, Set<String> inputKeys) {
			this.cache = cache;
			this.inputKeys = new HashSet<String>(inputKeys);
		}
		
	}
	
	@Listener
	public static class EntryChangeListener {

		@CacheEntriesEvicted
		public void observeEvicted(
				CacheEntriesEvictedEvent<String, String> event) {
			final String methodName = "observeEvicted";
			if (!event.isPre()) {
				Map<String, String> entries = event.getEntries();

				Cache<String, String> cache = event.getCache();
				for (Entry<String, String> entry : entries.entrySet()) {
					debug(methodName, String.format(
							"  next entry: key=%s value=%s", entry.getKey(),
							entry.getValue()));
				}
			}
		}
				
		@CacheEntryCreated
		public void observeAdd(CacheEntryCreatedEvent<String, String> event) {
			if (!event.isPre()) { // So that message is only logged after operation succeeded
				debug("observeAdd", "" + event);
			}
		}

		@CacheEntryRemoved
		public void observeRemove(CacheEntryRemovedEvent<String, String> event) {
			debug("observeRemove", "" + event);
		}

		@CacheEntryModified
		public void observeModified(CacheEntryModifiedEvent<String, String> event) {
			if (!event.isPre()) {
				debug("observeModified", "" + event);
			}
		}
	}
	
	private static class IOUtils {
		static InputStreamReader istream;
		static BufferedReader bufRead;

		static {
			istream = new InputStreamReader(System.in);
			bufRead = new BufferedReader(istream);
		}

		public static String readLine(String s) {
			System.out.print(s);
			String returnval = null;
			try {
				returnval = bufRead.readLine();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			return returnval;
		}

		public static void dumpWelcomeMessage() {
			System.out.println("Interactive test");
			System.out.println("=====================");
			System.out.println("Commands: " + Arrays.asList(COMMAND.values()));

		}
	}

	public static void main(String[] args) throws Exception {
		
		if ( args != null && args.length > 0 ) {
			for (String next : args ) {
				if ( next.contains("-listener") ) {
					useListener = true;
					break;
				}
			}
		}
		
		SimpleInteractiveTest test = null;
		try {
			test = new SimpleInteractiveTest();
			test.start();
		}  finally {
			test.terminate();
		}
	}
}
